function [T,leafInd,currentErr,ept] = MOUSSE(T,leafInd,vt,obsInd,method,epsilon0,mu,alpha,eta)
%----input
% 8 compulsory:
% T: initialized tree structre, by iterative bipartition
% leafInd: index of leaf node
% vt: current observation, a D by 1 vector
% obsInd: mark the observed entries with 1 and others 0
% epsilon0: when relative error is larger than this, consider split, else, consider merge
% mu: weight for tree structure complexity penalty
% alpha: forgetting factor, typically around 0.9
% 1 omissible:
% eta: maximum stepsize if using GROUSE subspace updating. Otherwise, not needed

%----output
% T: updated tree structure
% leafInd: updated index of leaf node
% currentErr: current observation's residual after projecting to the closest leaf node, 
% same as e_t in paper
% ept: closest leaf's weighted time averaging error, same as (D-d)*delta in paper

if nargin==8
    eta = 0;
end

[T,closestLeaf,leftORright,currentErr] = updateNodes(T,leafInd,vt,obsInd,method,alpha,eta);
ept = T(closestLeaf).error;

% see if split or merge is needed
leafErrTemp = T(closestLeaf).error;
virtualErr = T(leftORright).error;

mergeFlag = 1;
if length(leafInd)==1 % only one root, no way to merge
    mergeFlag = 0;
else
    myFather = T(closestLeaf).father; % find father
    fatherErr = T(myFather).error;
    % find sibling
    if T(myFather).left == closestLeaf
        sib = T(myFather).right;% sibling of thisLeaf
    else
        sib = T(myFather).left;
    end
    
    if T(T(sib).left).left~=-inf % if sibling is not a leaf node, don't merge
        mergeFlag = 0;
    end
end

if leafErrTemp>epsilon0 % consider split
    if leafErrTemp + mu*length(leafInd) > virtualErr + mu*(length(leafInd)+1)
        leafInd(leafInd==closestLeaf) = [];
        leftVirtual = T(closestLeaf).left;
        rightVirtual = T(closestLeaf).right;
        leafInd = [leafInd,leftVirtual,rightVirtual];
        
        % generate left for former leftVirtual
        T(end+1) = T(leftVirtual);
        [maxLen,maxInd] = max(T(leftVirtual).spread);
        majorAxis = T(leftVirtual).basis(:,maxInd);
        T(end).center = T(leftVirtual).center + sqrt(maxLen)/2  *majorAxis;
        T(end).spread(maxInd) = maxLen/4;
        T(end).left = -inf;
        T(end).right = -inf;
        T(end).father = leftVirtual;
        T(leftVirtual).left = length(T);
        % generate right for former leftVirtual
        T(end+1) = T(leftVirtual);
        T(end).center = T(leftVirtual).center - sqrt(maxLen)/2  *majorAxis;
        T(end).spread(maxInd) = maxLen/4;
        T(end).left = -inf;
        T(end).right = -inf;
        T(end).father = leftVirtual;
        T(leftVirtual).right = length(T);
        % generate left for former rightVirtual
        T(end+1) = T(rightVirtual);
        [maxLen,maxInd] = max(T(rightVirtual).spread);
        majorAxis = T(rightVirtual).basis(:,maxInd);
        T(end).center = T(rightVirtual).center + sqrt(maxLen)/2  *majorAxis;
        T(end).spread(maxInd) = maxLen/4;
        T(end).left = -inf;
        T(end).right = -inf;
        T(end).father = rightVirtual;
        T(rightVirtual).left = length(T);
        % generate right for former rightVirtual
        T(end+1) = T(rightVirtual);
        T(end).center = T(rightVirtual).center - sqrt(maxLen)/2  *majorAxis;
        T(end).spread(maxInd) = maxInd/4;
        T(end).left = -inf;
        T(end).right = -inf;
        T(end).father = rightVirtual;
        T(rightVirtual).right = length(T);
        % update activeNode info
        %activeNode = [activeNode,length(T)-3:length(T)];
        
        %[~,newFitErr] = dist2Node(T(leftORright),vt);
        %error(end) = error(t)*lambda + newFitErr;
        %error(end) = T(leftORright).relativeErr;
        
        %%%% current point error should be modified since the split
%         centerSub = vt(obsInd)-T(leftORright).center(obsInd);
%         U = T(leftORright).basis;
%         w = U(obsInd,:)\centerSub;
%         r = centerSub-U(obsInd,:)*w;
%         currentErr = norm(r).^2;
        
        ept = T(leftORright).error;
        %%%% current point error should be modified since the split
    end
    
else
    if (mergeFlag==1) && leafErrTemp + mu*length(leafInd) > fatherErr + mu*(length(leafInd)-1)
        %(max([leafErrTemp,sibErr]) + mu*log(length(leafInd)) > fatherErr + mu*log(length(leafInd)-1))
        %if leafErr + sibErr + mu*log(length(leafInd)) > fatherErr + mu*log(length(leafInd)-1)
        
        %             activeNode = setdiff(activeNode,[T(closestLeaf).left,T(closestLeaf).right,...
        %                 T(sib).left,T(sib).right]);
        
        
        
        
        %%%% current point error should be modified since the merge
        
%         centerSub = vt(obsInd)-T(myFather).center(obsInd);
%         U = T(myFather).basis;
%         w = U(obsInd,:)\centerSub;
%         r = centerSub-U(obsInd,:)*w;
%         currentErr = norm(r).^2;
        
        
        
        ept = T(myFather).error;
        %%%% current point error should be modified since the merge
        
        
        leafInd = setdiff(leafInd,[closestLeaf,sib]);
        leafInd = [leafInd, myFather];
        
        del1 = T(closestLeaf).left;
        del2 = T(closestLeaf).right;
        del3 = T(sib).left;
        del4 = T(sib).right;
        del = [del1 del2 del3 del4];
        
        for i=setdiff(1:length(T),del)
            if ismember(T(i).left,del)
                T(i).left = -inf;
            else
                T(i).left = T(i).left-sum(T(i).left>del);
            end
            
            if ismember(T(i).right,del)
                T(i).right = -inf;
            else
                T(i).right = T(i).right-sum(T(i).right>del);
            end
            
            T(i).father = T(i).father-sum(T(i).father>del);
            
        end
        
        for i=1:length(leafInd)
            leafInd(i) = leafInd(i)-sum(leafInd(i)>del);
        end
        T(del) = [];
        
        
        %error(end) = fatherErr;
    end
    
end

